/* 

     minrad.c : This file contains a single procedure that computes the
                "minrad" of a link L (in the sense of Eric Rawdon) and 
                returns it as a double.    

                We may later fold this into another source file.
                Note that we use some of the linear algebra stuff
                from vector.c. 

  $Id: minrad.c,v 1.13 2004/10/08 00:15:20 cantarel Exp $

*/

/* Copyright 2004 The University of Georgia. */

/* This file is part of liboctrope.
   
liboctrope is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

liboctrope is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with liboctrope; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif
#include "octrope.h"


double octrope_minrad(const octrope_link *L, const double epsilon,
                      octrope_mrloc *min_rad_locs,
                      int mr_size, int *num_min_rad_locs)
{
  int i,j,k,l;
  double mr = {DBL_MAX}, alpha, rad;
  octrope_vector in,out;
  double normin, normout;
  double rplus, rminus, this_mr;
  int start,end;

  /* We start with some initializations. */
  
  octrope_error_num = octrope_error_str[0] = 0;
  octrope_link_fix_wrap(L);
  if (num_min_rad_locs != NULL) *num_min_rad_locs = 0;
  
  /* Now we work. */

  for(i=0;i<L->nc;i++) {                  /* Loop over components. */

    if (L->cp[i].acyclic) {
      start = 1;
      end = L->cp[i].nv - 1;
      out = octrope_vminus(L->cp[i].vt[1],L->cp[i].vt[0]);      
      normout = octrope_norm(out); 
    } else {
      start = 0;
      end = L->cp[i].nv;
      out = octrope_vminus(L->cp[i].vt[0],L->cp[i].vt[-1]);      
      normout = octrope_norm(out); 
    }

    /* Now we handle the main loop. */

    for (j = start; j < end; j++) {
      
      in     = out;                   /* Pass back "outgoing" data. */
      normin = normout;               
      
      out = octrope_vminus(L->cp[i].vt[j+1],L->cp[i].vt[j]);
      normout = octrope_norm(out);
      
      alpha = acos(octrope_dot(in,out)/(normin*normout));
      rad   = 1/(2*tan(alpha/2));
      
      rminus = normin*rad;
      rplus  = normout*rad;

      this_mr = (rminus < rplus) ? rminus : rplus;

      mr = (mr < this_mr) ? mr : this_mr;

      if (num_min_rad_locs != NULL) { /* We are recording locations. */
        if (this_mr <= mr) { /* This one is of interest. */
          if (*num_min_rad_locs < mr_size) { /* And we have room. */
            min_rad_locs[*num_min_rad_locs].component = i;
            min_rad_locs[*num_min_rad_locs].vert = j;
            min_rad_locs[*num_min_rad_locs].mr = this_mr;
  
            (*num_min_rad_locs)++;
          } else { /* Or we can create room. */ 
            /* Note that be the way we built the list, all of the minrads 
               with value <= mr+epsilon will be at the end of the list. */
            k = 0;
            while ((k < *num_min_rad_locs) &&
                   (min_rad_locs[k].mr > mr+epsilon)) { k++; }
            if (k > 0) { /* There are some slots we can free up */
              l = k; /* distance to move things */
              while (k < *num_min_rad_locs) {
                min_rad_locs[k-l] = min_rad_locs[k];
                k++;
              }
              (*num_min_rad_locs) -= l;
  
              /* Now slide in the new one */
              min_rad_locs[*num_min_rad_locs].component = i;
              min_rad_locs[*num_min_rad_locs].vert = j;
              min_rad_locs[*num_min_rad_locs].mr = this_mr;
  
              (*num_min_rad_locs)++;
            }
          }
        }
      }
      
#ifdef DEBUG
      if (octrope_debug_level() > 8) {
        printf("%d/%d & %d/%d:\n",j,j-1,j+1,j);
        printf("  (%3.3f,%3.3f,%3.3f)-(%3.3f,%3.3f,%3.3f)=(%3.3f,%3.3f,%3.3f) |%3.3f|\n",
          L->cp[i].vt[j].c[0], L->cp[i].vt[j].c[1], L->cp[i].vt[j].c[2],
          L->cp[i].vt[j-1].c[0], L->cp[i].vt[j-1].c[1], L->cp[i].vt[j-1].c[2],
          in.c[0],in.c[1],in.c[2],normin);
        printf("  (%3.3f,%3.3f,%3.3f)-(%3.3f,%3.3f,%3.3f)=(%3.3f,%3.3f,%3.3f) |%3.3f|\n",
          L->cp[i].vt[j+1].c[0], L->cp[i].vt[j+1].c[1], L->cp[i].vt[j+1].c[2],
          L->cp[i].vt[j].c[0], L->cp[i].vt[j].c[1], L->cp[i].vt[j].c[2],
          out.c[0],out.c[1],out.c[2],normout);
        printf("  alpha: %3.3f, rad: %3.3f, rminus: %3.3f, rplus: %3.3f, mr: %3.3f\n",
          alpha,rad,rminus,rplus,mr);
      }
#endif /* DEBUG */
    }
    
  }

  /* Pare down the list to be only the mr's within epsilon of minimum */
  if (num_min_rad_locs != NULL) {
    /* Note that by the way we built the list, all of the minrads 
       with value <= mr+epsilon will be at the end of the list. */
    k = 0;
    while ((k < *num_min_rad_locs) && 
           (min_rad_locs[k].mr > mr+epsilon)) { k++; }
    if (k > 0) { /* There are some slots we can free up */
      l = k; /* distance to move things */
      while (k < *num_min_rad_locs) {
        min_rad_locs[k-l] = min_rad_locs[k];
        k++;
      }
      (*num_min_rad_locs) -= l;
    }
  }
  
  return mr;
}
